<?php
/**
 * Handler: to create an easy way to build urls for using Imengine API.
 */

namespace EwTools\Handler;

use RuntimeException;

/**
 * Imengine
 *
 * @link    http://infomaker.se
 * @package EwTools\Handler
 * @since   EwTools\Handler\Imengine 0.1
 */
class Imengine {
    
    /**
     * Contains the base url to the Imengine server
     *
     * @var string
     */
    private $base_url;
    
    /**
     * Contains an array of the parameters that will be sent to Imengine server
     *
     * @var array
     */
    private $params;
    
    /**
     * Contains the default parameters that's always sent to Imengine regardless of the function used
     *
     * @var array
     */
    private static $default_params = [
        'function' => null,
        'type'     => 'preview', // preview|primary
        'source'   => 'false',
        'q'        => 75,
    ];
    
    /**
     * Imengine constructor.
     *
     * @since 0.1
     *
     * @param string $imengine_url
     * @param array  $params
     */
    private function __construct( $imengine_url, array $params = [] ) {
        $this->base_url = $imengine_url;
        $this->params   = array_replace( static::$default_params, $params );
    }
    
    /**
     * Use an uuid to specify witch image to manipulate.
     * (The image must exist in the Open content configured in the Imengine api in use)
     *
     * @param string $uuid
     *
     * @since 0.1
     * @return null|string
     */
    public function fromUuid( $uuid ) {
        return $this->buildFromParams( array_merge( [ 'uuid' => $uuid ], $this->params ) );
    }
    
    /**
     * Use an external image to manipulate.
     * (The new version will be saved on Imengine server)
     *
     * @param string $imgsrc
     *
     * @since 0.1
     * @return null|string
     */
    public function fromSrc( $imgsrc ) {
        
        // Add format if not already set
        if( ! isset( $this->params[ 'format' ] ) ) {
            $this->params[ 'format' ] = pathinfo( strtok( $imgsrc, '?' ), PATHINFO_EXTENSION );
        }
        
        return $this->buildFromParams( array_merge( [ 'imgsrc' => trim( $imgsrc ) ], $this->params ) );
    }
    
    // Chainable functions: functions to chain on mirror functions to further manipulate your query before sending it to Imengine
    // ======================================================
    
    /**
     * Add or override parameters not specific to the function being used
     *
     * @param array $params
     *
     * @since 0.1
     * @return Imengine
     */
    public function with( array $params = [] ) {
        $this->params = array_replace( $this->params, $params );
        
        return $this;
    }
    
    /**
     * Function to allow for quality change of the requested image
     *
     * @param int $quality [1-100]
     *
     * @since 0.1
     * @return Imengine
     */
    public function quality( $quality ) {
        if( is_int( $quality ) && $quality > 0 && $quality <= 100 ) {
            $this->params = array_replace( $this->params, [ 'q' => $quality ] );
        }
        
        return $this;
    }
    
    // Protected functions:
    // ======================================================
    
    /**
     * Function to build the query url to Imengine API
     *
     * @param array $params
     *
     * @since 0.1
     * @return string
     */
    protected function buildFromParams( array $params = [] ) {
        return "{$this->base_url}?" . http_build_query( $params );
    }
    
    /**
     * Function to calculate height from given width using ration
     *
     * @param int    $width
     * @param string $ratio
     *
     * @since 0.1
     * @return int
     */
    protected static function calculateRatioHeight( $width, $ratio ) {
        list( $ratio_width, $ratio_height ) = explode( ':', $ratio );
        
        return (int)round( ( $width / (int)$ratio_width ) * (int)$ratio_height );
    }
    
    /**
     * Function to create an instance with only params.
     * Used by all Imengine server-function mirrored functions
     *
     * @param string $func
     * @param array  $params
     *
     * @since 0.1
     * @return Imengine
     */
    public static function createMirrorFunction( $func, array $params = [] ) {
        try {
            $base_url = static::getImengineUrl();
        } catch ( RuntimeException $e ) {
            error_log( $e->getMessage() );
            $base_url = '/';
        }
        return new static( $base_url, array_merge( [ 'function' => $func ], $params ) );
    }
    
    /**
     * Function to get DEV or PROD img-base URL (with or without CDN)
     *
     * @since 0.1
     * @return string url
     * @throws \RuntimeException
     */
    protected static function getImengineUrl() {
        if( ! defined( 'CF_IMENGINE' ) ) {
            throw new RuntimeException( 'CF_IMENGINE is not defined and can therefor not serve images' );
        }
        
        return defined( 'CF_IMENGINE' ) ? CF_IMENGINE : false;
    }
    
    // Mirror functions: mirrors the functions found on Imengine API
    // ======================================================
    
    /**
     * Scale image to cover the limits.
     * (Can exceed limits)
     *
     * @param int $width
     * @param int $height
     *
     * @since 0.1
     * @return Imengine
     * @throws \RuntimeException
     */
    public static function cover( $width = null, $height = null ) {
        return static::createMirrorFunction( 'cover', compact( [ 'width', 'height' ] ) );
    }
    
    /**
     * Function to create a url from scratch
     *
     * @param array $params
     *
     * @since 0.1
     * @return Imengine
     * @throws \RuntimeException
     */
    public static function create( array $params = [] ) {
        try {
            $base_url = static::getImengineUrl();
        } catch ( RuntimeException $e ) {
            error_log( $e->getMessage() );
            $base_url = '/';
        }
        return new static( $base_url, $params );
    }
    
    /**
     * Crop image into defined crop-size from custom start-positions.
     * Then scale into defined size
     *
     * @param int $crop_w
     * @param int $crop_h
     * @param int $width
     * @param int $height
     * @param int $x
     * @param int $y
     *
     * @since 0.1
     * @return Imengine
     * @throws \RuntimeException
     */
    public static function cropResize( $crop_w, $crop_h, $width = null, $height = null, $x = 0, $y = 0 ) {
        return static::createMirrorFunction( 'cropresize', compact( [
            'crop_w',
            'crop_h',
            'x',
            'y',
            'width',
            'height'
        ] ) );
    }
    
    /**
     * Scale image to fit inside the limits.
     * Add scale up to make the image cover the maxsize-area
     * (Can exceed limits on scale up)
     *
     * @param int  $maxsize
     * @param bool $scaleup
     *
     * @since 0.1
     * @return Imengine
     * @throws \RuntimeException
     */
    public static function fit( $maxsize, $scaleup = true ) {
        return static::createMirrorFunction( 'fit', compact( [ 'maxsize', 'scaleup' ] ) );
    }
    
    /**
     * Crop image into defined size from center position
     *
     * @param int $width
     * @param int $height
     *
     * @since 0.1
     * @return Imengine
     */
    public static function hardcrop( $width, $height ) {
        return static::createMirrorFunction( 'hardcrop', compact( [ 'width', 'height' ] ) );
    }
    
    /**
     * Crop image into defined size from custom start-positions.
     *
     * @param int $width
     * @param int $height
     * @param int $x
     * @param int $y
     * @param int $zoom
     *
     * @since 0.1
     * @return Imengine
     */
    public static function npCrop( $width, $height, $x = 0, $y = 0, $zoom = 1 ) {
        return static::createMirrorFunction( 'np_crop', compact( [
            'width',
            'height',
            'x',
            'y',
            'zoom'
        ] ) );
    }
    
    /**
     * Fetch image with original size
     *
     * @since 0.1
     * @return Imengine
     */
    public static function original() {
        return static::createMirrorFunction( 'original' );
    }
    
    /**
     * Crop image into defined size from center position
     *
     * @param int $width
     * @param int $height
     *
     * @since 0.1
     * @return Imengine
     */
    public static function thumbnail( $width, $height ) {
        return static::createMirrorFunction( 'thumbnail', compact( [ 'width', 'height' ] ) );
    }
    
    // Extended mirror functions: functions with extended functionality och the basic mirror functions
    // ======================================================
    
    /**
     * Crop image into given ratio
     *
     * @param $width
     * @param $ratio
     *
     * @since 0.1
     * @return Imengine
     */
    public static function ratio( $width, $ratio = '3:2' ) {
        return static::hardcrop( $width, static::calculateRatioHeight( $width, $ratio ) );
    }
    
    /**
     * Crop image into given ratio
     *
     * @param int    $original_width  Original image width
     * @param int    $original_height Original image height
     * @param int    $width           Width to resize image to
     * @param string $ratio
     *
     * @since 0.1
     * @return Imengine
     */
    public static function autoCrop( $original_width, $original_height, $width, $ratio = '3:2' ) {
        
        try {
            list( $diff_width, $diff_height ) = static::calculateCropValues( (int)$original_width, (int)$original_height, $ratio );
            
            return static::createMirrorFunction( 'cropresize', [
                'crop_w' => static::formatFloatCropNumber( $diff_width ),
                'crop_h' => static::formatFloatCropNumber( $diff_height ),
                'x'      => static::formatFloatCropNumber( ( 1 - $diff_width ) / 2 ),
                'y'      => static::formatFloatCropNumber( ( 1 - $diff_height ) / 2 ),
                'width'  => $width,
                'height' => static::calculateRatioHeight( $width, $ratio )
            ] );
            
        } catch ( RuntimeException $e ) {
            
        }
        
    }
    
    /**
     * Format image crop number
     *
     * @param $number
     *
     * @return float
     */
    private static function formatFloatCropNumber( $number ) {
        $number = (float)sprintf( '%.5f', $number + 0 );
        if( strlen( $number ) === 1 ) {
            if( $number == 1 ) {
                $number = 0.99999;
            } else if( $number == 0 ) {
                $number = 0.00001;
            }
        }
        
        return $number;
    }
    
    /**
     * Calculate the new values in percentage to the original with maximum sizing
     *
     * @param $original_width
     * @param $original_height
     * @param $ratio
     *
     * @since 1.0.0
     * @return array
     */
    private static function calculateCropValues( $original_width, $original_height, $ratio ) {
        
        list( $ratio_width, $ratio_height ) = explode( ':', $ratio );
        $original_ratio = $original_width / $original_height;
        $new_ratio      = $ratio_width / $ratio_height;
        
        if( $original_ratio > $new_ratio ) {
            $new_height = $original_height;
            $new_width  = $new_height * $new_ratio;
            
            $diff_height = $new_height / $original_height;
            $diff_width  = $new_width / $original_width;
            
        } else {
            $new_width  = $original_width;
            $new_height = $new_width / $new_ratio;
            
            $diff_height = $new_height / $original_height;
            $diff_width  = $new_width / $original_width;
        }
        
        return [
            $diff_width,
            $diff_height,
        ];
    }
    
    /**
     * Function for fetching the Raw format of the image (May have an unlimited size)
     *
     * @param string $uuid
     * @param int    $quality
     *
     * @since 0.1
     * @return string
     */
    public static function raw( $uuid, $quality = 100 ) {
        try {
            $base_url = static::getImengineUrl();
        } catch ( RuntimeException $e ) {
            error_log( $e->getMessage() );
            $base_url = '/';
        }
        return ( new static( $base_url, [ 'type' => 'primary', 'q' => $quality ] ) )->fromUuid( $uuid );
    }
    
    /**
     * Crop image into a square
     *
     * @param $size
     *
     * @since 0.1
     * @return Imengine
     */
    public static function square( $size ) {
        return static::hardcrop( $size, $size );
    }
}
