<?php
/**
 * Helper class for setting up Twig with functions, filters and loaders
 */

namespace EwTools\Twig;

use EwTools\Handler\Sidebars;
use EwTools\Models\SettingsParameter;
use Twig_SimpleFilter;
use Twig_SimpleFunction;

/**
 * ViewSetup
 *
 * @link    http://infomaker.se
 * @package EwTools\Twig
 * @since   EwTools\Twig\ViewSetup 0.1
 */
abstract class ViewSetup {
    
    /**
     * Whether to run Twig in cache state
     *
     * @var bool
     */
    protected $cache = false;
    
    /**
     * Time in seconds to cache the output if cache is set to true
     *
     * @var int
     */
    protected $cache_time = 86400;
    
    /**
     * Whether to run Twig in debug state
     *
     * @var bool
     */
    protected $debug = true;
    
    /**
     * Contains Wordpress functions that will be added to the list off functions
     * that will be available in the templates
     * function: [func_name] => [custom_name]
     *
     * @var array
     */
    protected $wp_functions = [
        'get_header'                   => '',
        '__'                           => 'translate',
        'get_bloginfo'                 => '',
        'get_footer'                   => '',
        'get_the_title'                => '',
        'the_content'                  => '',
        'body_class'                   => '',
        'site_url'                     => '',
        'self_link'                    => '',
        'get_permalink'                => '',
        'wp_head'                      => '',
        'wp_footer'                    => '',
        'get_template_directory_uri'   => '',
        'get_stylesheet_directory_uri' => ''
    ];
    
    /**
     * Contains the registered filters available in the templates
     *
     * @var array
     */
    private $filters = [];
    
    /**
     * Contains the functions available in the templates
     *
     * @var array
     */
    private $functions = [];
    
    /**
     * Contains the global variables available in the templates
     *
     * @var array
     */
    private $globals = [];
    
    /**
     * Contains an instance of the loader-class Twig-template folders
     *
     * @var FilesystemLoader
     */
    private $loader;
    
    /**
     * ViewSetup constructor.
     *
     * @since 0.1
     */
    public function __construct() {
        
        $this->registerTwigFolder( 'tools', __DIR__ . '/templates' );
        
        // Set Default functions
        $this->addFunction( 'render_partial', [
            View::class,
            'render'
        ] )->addFunction( 'render_sidebar', [
            Sidebars::class,
            'render'
        ] )->addFunction( 'settings_parameter', [
            SettingsParameter::class,
            'getValue'
        ] )->addFunction( 'class_list', function ( $classes ) {
            return implode( ' ', array_filter( (array)$classes ) );
        } );
        
        // Wordpress functions
        // ======================================================
        foreach ( $this->wp_functions as $wp_function => $name ) {
            $this->addWpFunction( $wp_function, $name );
        }
        
        $this->init();
    }
    
    /**
     * Setup your Twig-environment with settings folders to be loaded, states etc.
     *
     * @since 0.1
     * @return void
     */
    abstract public function setupTwig();
    
    /**
     * Help function for adding template-functions to twig
     *
     * @param string   $name
     * @param callable $callable
     * @param array    $options
     *
     * @since 0.1
     * @return $this
     */
    protected function addFunction( $name, $callable, array $options = [] ) {
        $this->functions[] = new Twig_SimpleFunction( $name, $callable, $options );
        
        return $this;
    }
    
    /**
     * Help function for adding template-functions to twig
     *
     * @param string   $name
     * @param callable $callable
     * @param array    $options
     *
     * @since 0.1
     * @return $this
     */
    protected function addFilter( $name, $callable, array $options = [] ) {
        $this->filters[] = new Twig_SimpleFilter( $name, $callable, $options );
        
        return $this;
    }
    
    /**
     * @param string $key
     * @param mixed  $value
     *
     * @since 1.0.0
     * @return void
     */
    protected function addGlobal( $key, $value ) {
        $this->globals[ $key ] = $value;
    }
    
    /**
     *  Function for adding wordpress function as template-functions to twig
     *
     * @param string $func_name
     * @param string $name
     *
     * @since 0.1
     * @return static
     */
    protected function addWpFunction( $func_name, $name = null ) {
        $name = ( is_string( $name ) && ! empty( $name ) ) ? $name : $func_name;
        $this->addFunction( $name, $func_name );
        
        return $this;
    }
    
    /**
     *
     *
     * @since 1.0.0
     * @return FilesystemLoader
     */
    protected function getLoader() {
        if( null === $this->loader ) {
            $this->loader = new FilesystemLoader();
        }
        
        return $this->loader;
    }
    
    /**
     * Function for adding folders containing the Twig-files to be loaded
     *
     * @param string $namespace
     * @param string $folder_path
     * @param array  $exclude
     *
     * @throws \Twig_Error
     * @return $this
     * @since 0.1
     */
    protected function registerTwigFolder( $namespace, $folder_path, array $exclude = [] ) {
        if( is_dir( $folder_path ) ) {
            $this->getLoader()->addPathAndSubfolders( $folder_path, $namespace, $exclude );
        }
        
        return $this;
    }
    
    /**
     * Initiate
     *
     * @param array $options
     *
     * @since 0.1
     * @return void
     */
    protected function init( array $options = [] ) {
        
        // Use the abstract class in child-classes to customize setup
        $this->setupTwig();
        
        $twig = Every_Twig::get_instance()->init( [ $this->getLoader() ], $this->functions, $this->filters, array_replace( [
            'debug'     => $this->debug,
            'cache'     => $this->cache,
            'cachetime' => $this->cache_time
        ], $options ) );
        
        //Add global device variable
        if( defined( 'EVERY_DEVICE' ) ) {
            $this->addGlobal( 'device', EVERY_DEVICE );
        }
        
        foreach ( $this->globals as $key => $value ) {
            $twig->addGlobal( $key, $value );
        }
    }
    
    /**
     * Setup Twig views
     *
     * @since 0.1
     * @return static
     */
    public static function setup() {
        return new static();
    }
}